/*
 * prefetcher.c -- F-CPU instruction-level emulator Instruction Cache L0 methods
 * Copyright (C) 2003 Pierre Tardy <tardyp@free.fr>
 * Copyright (C) 2003 Franois Vieville <vieville@poste.isima.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include "emu.h"

int lineRef[NUM_REG];
BOOL waitPrefetcher = false;

#define NUM_PREFETCH_LINES 30
typedef struct s_PrefetchLine
{
  int regRef;
  U64 addr;
  U32 instr;
  BOOL valid;
}PrefetchLine;

PrefetchLine lines[NUM_PREFETCH_LINES];

unsigned int PCRef = 0;
U64 PC = 0;
unsigned int PrefetchPC = 0;
int jumping = 0;

static void pf_invalidate (int linenb)
{
  lines[linenb].valid = false;

  if(lines[linenb].regRef != 0 && lineRef[lines[linenb].regRef]== linenb)
    lineRef[lines[linenb].regRef]= 0xff;// we just forget this register
  lines[linenb].regRef=0;

}
// this process is called at each clock rising edge
U32 pf_getNextInstruction()
{
  if(lines[PCRef].addr != PC || !lines[PCRef].valid)
    {  // we ve got a miss!
      waitPrefetcher = true; // signal that say the core that data is not ready. core: please sleep!
      PrefetchPC = PCRef;
      lines[PCRef].addr = PC;
      
      if(!jumping) // dont erase the register reference if we are jumping and we dont have prefetched yet.
	pf_invalidate(PCRef);
    }
  else
    { 
      U32 inst;
      // we have the data, return it.
      waitPrefetcher = false;
      PC = PC + 4;
      inst = lines[PCRef].instr;
      PCRef = (PCRef+1)  % NUM_PREFETCH_LINES;
      jumping = 0;
      return inst;
    }
  return 0;
}
// this process is called when the decoder finds a loadAddri. addr is on the cross bar, 
// regnum is passed from decoder to prefetcher
void pf_loadaddr(int regnum, U64 addr)
{
  // set the prefetcher to prefetch this addr
  lineRef[regnum] = PrefetchPC;
  lines[PrefetchPC].regRef = regnum;
  lines[PrefetchPC].addr = addr;
  lines[PrefetchPC].valid = false;
}
// this process is called when the decoder finds an instruction that changes Ri, 
// regnum is passed from decoder to prefetcher
void pf_RiChanged(int regnum) // some instruction other than loadAddri that change Ri
{
  lineRef[regnum] = 0xff;
}
// this process is called when we have to jump to a given register @, 
void pf_doJmp(int regnum)
{
  if(lineRef[regnum] != 0xff && lines[lineRef[regnum]].regRef == regnum) // we have it in the cache!
    {
      
      PCRef = lineRef[regnum];
      PC = lines[PCRef].addr;
      fprintf(stderr,"jmp that hit!\n");
      // Ok, lets wait next cycle.
    }
  else // damn! its a miss!
    {
      fprintf(stderr,"jmp that missed!\n");
      // ok, abandon what we were prefetching..
      PCRef = PrefetchPC;
      lineRef[regnum] = PrefetchPC;
      lines[PrefetchPC].regRef = regnum;
      lines[PrefetchPC].addr = PC = regs.r[regnum].o[0];// maybe this is on the crossbar but I fear we'll have to wait one cycle to have this value
      lines[PrefetchPC].valid = false;
      jumping = 1;

      // the next getNextInstruction will fall into a miss.
    }
  
}
// this is called by the emulator debugger when PC is forced
void pf_updatePC()
{
  lines[PrefetchPC].regRef = 0;
  lines[PrefetchPC].addr = regs.r_pc.C(o,0);
  pf_invalidate(PrefetchPC);
  
}
// this process is called at each clock rising edge
void pf_doPrefetch()
{
  // L1DoLoadMem will return true at the nth times it has been called if its passed addr do not change
  // this is to emulate the slowness of the L1
  if(L1DoLoadMem(lines[PrefetchPC].addr)) // return 1 if finished
    {
      U64 lastPC = PrefetchPC;
      lines[PrefetchPC].instr = getL1Data();
      lines[PrefetchPC].valid = true;
      // calculate the next opcode to prefetch
      PrefetchPC++;
      PrefetchPC = PrefetchPC  % NUM_PREFETCH_LINES; // loop

      // destroy the last recently used line
      pf_invalidate(PrefetchPC);
      
      // calculate the next prefetch addr.
      lines[PrefetchPC].addr = lines[lastPC].addr+4;

    }  
}

void pf_showPrefetchLines()
{
  int i;
  fprintf(stderr,"|Ri|lin| |lin|Ri|     addr       |  instr |valid\n");
  for( i = 0 ; i < /*NUM_PREFETCH_LINES*/30; i++)
    {
      if(i < NUM_REG)
	fprintf(stderr,"|%02d %03X| ", i,lineRef[i]);
      else
	printf("---------");
      fprintf(stderr,"|%03X %02d %016X %08X %s|%s%s%s\n",
	     i,
	     lines[i].regRef,
	     (U32)lines[i].addr,
	     lines[i].instr,
	     ((lines[i].valid)?"+":"-"),
	     ((PrefetchPC == i)?"PrefetchPC ":""),
	     ((PCRef == i)?"PCRef ":""),
	     ((PC == lines[i].addr)?"PC ":"")
	     );
    }
}
